/* The calling syntax is:

            [v] = oipmc(u, setpar)

*/

#include <math.h>
#include "cmex.h"

/* Input Arguments */
#define u_in       prhs[0]
#define setpar_in  prhs[1]
#define opt_in     prhs[2]

/* Output Arguments */
#define v_out    plhs[0]

void mexFunction(nlhs, plhs, nrhs, prhs)
int nlhs, nrhs;
Matrix *plhs[], *prhs[];
{
  double *v, *u, *setpar,*optv;
  int it,opt,antit;

  if (nrhs!=3) mexErrMsgTxt("Hough requires three input arguments.");
  if (nlhs>1) mexErrMsgTxt("Hough returns only one output arguments.");

  if ( mxGetM(u_in) == 0 || mxGetN(u_in) == 0 ) 
    mexErrMsgTxt("Input matrix of size zero..");

  /* Assign pointers to the parameters */
  u      = mxGetPr(u_in);
  setpar = mxGetPr(setpar_in);
  optv   = mxGetPr(opt_in);
  opt    = (int)optv[0];
  antit  = (int)optv[1];
 
  /* Create a matrix for the return argument */
  v_out = mxCreateFull((int)setpar[3],(int)setpar[5],REAL);

  /* Assign pointers to the output parameters */
  v = mxGetPr(v_out);

  /* Do the actual computations in a subroutine */
  for (it=0;it<antit;it++)
    switch (opt) {
      case 1 :
        oHoughO1(v,u,setpar);
        break;
      case 2 :
        oHoughO2(v,u,setpar);
        break;
      case 3 :
        oHoughO3(v,u,setpar);
        break;
      case 4 :
        oHoughO4(v,u,setpar);
        break;
      default :
        mexErrMsgTxt("Unknown Opt type");
      }
}

oHoughO4(double *v,double *u,double *setpar)
{

  int m,n, M;
  int r, R;
  int t, T,moffset,noffset;
  double sum,x_min,Delta_x,Delta_rho,rho_min,Delta_theta;
  double *costheta,*sintheta,theta,x,y;
  double rhooffset,theta_min;
  double sigval;
  double *xc,*ys;
  double cst;
  
  M            = setpar[0];
  x_min        = setpar[1];
  Delta_x      = setpar[2];
  T            = setpar[3];
  Delta_theta  = setpar[4];
  R            = setpar[5];
  rho_min      = setpar[6];
  Delta_rho    = setpar[7];
  theta_min    = setpar[8];

  for (t=0;t<R*T;t++)
    v[t] = 0.0;

  if (R<=Delta_x/Delta_rho*sqrt(2.0)*(M-1))
  {
    printf("Number of samples in the rho-direction R is to small\n");
  }
  else
  {
    costheta=(double *)malloc(T*sizeof(double));
    sintheta=(double *)malloc(T*sizeof(double));
    xc=(double *)malloc(M*T*sizeof(double));
    ys=(double *)malloc(M*T*sizeof(double));
    
    for (t=0;t<T;t++)
    {
      theta=t*Delta_theta;
      costheta[t]=cos(theta);
      sintheta[t]=sin(theta);
    }
    
    for (m=0;m<M;m++)
    {
      x=(x_min+m*Delta_x)/Delta_rho; /* x(m) = y(m) */    
      cst=rho_min/Delta_rho+0.5;
      for (t=0;t<T;t++)
      {
        xc[m*T+t]=x*costheta[t];
        ys[m*T+t]=x*sintheta[t]-cst;
      }
    }
    for (m=0;m<M;m++)
    {
      x=x_min+m*Delta_x;
      moffset=m*T;
      for (n=0;n<M;n++)
      {
        sigval=u[m+n*M];
        if (fabs(sigval)>1E-30)
        {
          y=x_min+n*Delta_x;
          noffset=n*T;
          for (t=0;t<T;t++)
            v[t+T*(int)(xc[moffset+t]+ys[noffset+t])]+=sigval;
        }
      }
    }
    free(costheta);
    free(sintheta);
    free(xc);
    free(ys);
  }
}

oHoughO3(double *v,double *u,double *setpar)
{

  int m,n, M;
  int r, R;
  int t, T;
  double sum,x_min,Delta_x,Delta_rho,rho_min,Delta_theta;
  double *costheta,*sintheta,theta,x,y;
  double rhooffset,theta_min;
  double sigval,offset;

  M            = setpar[0];
  x_min        = setpar[1];
  Delta_x      = setpar[2];
  T            = setpar[3];
  Delta_theta  = setpar[4];
  R            = setpar[5];
  rho_min      = setpar[6];
  Delta_rho    = setpar[7];
  theta_min    = setpar[8];

  for (t=0;t<R*T;t++)
    v[t] = 0.0;

  if (R<=Delta_x/Delta_rho*sqrt(2.0)*(M-1))
    mexErrMsgTxt("Number of samples in the rho-direction R is to small\n");
  else
  {
    costheta=(double *)malloc(T*sizeof(double));
    sintheta=(double *)malloc(T*sizeof(double));
    for (t=0;t<T;t++)
    {
      theta=t*Delta_theta;
      costheta[t]=cos(theta);
      sintheta[t]=sin(theta);
    }
    
    offset=0.5-rho_min/Delta_rho;
    for (m=0;m<M;m++)
    {
      x=(x_min+m*Delta_x)/Delta_rho;
      for (n=0;n<M;n++)
      {
        y=(x_min+n*Delta_x)/Delta_rho;
        sigval=u[m+n*M];
        if (fabs(sigval)>1E-30)
          for (t=0;t<T;t++)
            v[t+T*(int)(offset+x*costheta[t]+y*sintheta[t])]+=sigval;
      }
    }
    free(sintheta);
    free(costheta);
  }
}


oHoughO2(double *v,double *u,double *setpar)
{

  int m,n, M;
  int r, R;
  int t, T;
  double sum,x_min,Delta_x,Delta_rho,rho_min,Delta_theta;
  double *costheta,*sintheta,theta,x,y;
  double rhooffset,theta_min;
  double sigval,off;

  M            = setpar[0];
  x_min        = setpar[1];
  Delta_x      = setpar[2];
  T            = setpar[3];
  Delta_theta  = setpar[4];
  R            = setpar[5];
  rho_min      = setpar[6];
  Delta_rho    = setpar[7];
  theta_min    = setpar[8];

  costheta=(double *)malloc(T*sizeof(double));
  sintheta=(double *)malloc(T*sizeof(double));

  for (t=0;t<R*T;t++)
    v[t] = 0.0;

  for (t=0;t<T;t++)
  {
    theta=t*Delta_theta;
    costheta[t]=cos(theta);
    sintheta[t]=sin(theta);
  }

  off=0.5-rho_min/Delta_rho;
  for (m=0;m<M;m++)
  {
    x=(x_min+m*Delta_x)/Delta_rho;
    for (n=0;n<M;n++)
    {
      y=(x_min+n*Delta_x)/Delta_rho;
      sigval=u[m+n*M];
      if ((sigval>1E-30) || (sigval<-1E-30))
        for (t=0;t<T;t++)
        {
          r=(int)floor(off+x*costheta[t]+y*sintheta[t]);
          if ((r>=0) && (r<R)) v[t+r*T]+=sigval;
        }
    }
  }
  free(sintheta);
  free(costheta);
}


oHoughO1(double *v,double *u,double *setpar)
{

  int m,n, M;
  int r, R;
  int t, T;
  double sum,x_min,Delta_x,Delta_rho,rho_min,Delta_theta;
  double costheta,sintheta,theta,x,y;
  double rhooffset,theta_min;
  double sigval;

  M            = setpar[0];
  x_min        = setpar[1];
  Delta_x      = setpar[2];
  T            = setpar[3];
  Delta_theta  = setpar[4];
  R            = setpar[5];
  rho_min      = setpar[6];
  Delta_rho    = setpar[7];
  theta_min    = setpar[8];

  for (t=0;t<R*T;t++)
    v[t] = 0.0;

  for (m=0;m<M;m++)
  {
    x=x_min+m*Delta_x;
    for (n=0;n<M;n++)
    {
      y=x_min+n*Delta_x;
      sigval=u[m+n*M];
      if (fabs(sigval)>1E-30)
        for (t=0;t<T;t++)
        {
          theta=t*Delta_theta;
          r=(int)floor(0.5+(x*cos(theta)+y*sin(theta)-rho_min)/Delta_rho);
          if ((r>=0) && (r<R)) v[t+r*T]+=sigval;
        }
    }
  }
}




